﻿Imports System.Collections.Generic
Imports System.IO

Class PatternDatabase  ' Hash-keys can be saved in UInt32 <== max. 5x5 board with partitions of size max. 5 + 1 for empty tile => save 6 positions (0-15) => 6x4b=24b  
    Private type As PuzzleType
    Private rows As Integer, cols As Integer, parts As Integer
    Private partitions As Integer()()  ' [partition_index][tile_index] --> negative value = another partition,  values >= 0 = positions (indices) of tiles in hash key, value == parts = empty tile  
    Private partitioning As Integer()  ' indices into `partitions`  
    Private database As Dictionary(Of UInt32, Integer)()
    Private est_hash As UInt32()
    Public Const BITS_PER_POS As Integer = 4  ' 0 - 15  
    Public Const BIT_MASK As UInteger = 15  ' upper bound of the BITS_PER_POS range  

    Public Sub New(ByVal type As PuzzleType)
        Me.type = type
        InitDatabase()
        If type <> PuzzleType.PUZZLE_3x3 Then
            ' 3x3 puzzle will use online-calculated Manahattan distance,    
            ' because the lookup time is almost the same and there is no need of 
            ' PatternDB for puzzle of this size    
            If Not FileExists() Then
                For partition As Integer = 0 To parts - 1
                    BreadthFirstSearchGenerator(partition)
                    ' optional space-optimization: mirror configurations       
                    ' optional space-optimization: clean the database: remove all values <= node.Manahattan()       
                    ' -> don't use that, because I want to have all precomputed (it doesn't take so much of space)      
                    CleanDatabase(partition)
                Next
                SaveToFile()
            Else
                LoadFromFile()
            End If
        End If
    End Sub
 
    Private Function FileExists() As Boolean
        Dim filename As String = String.Format("puzzle_{0:d}x{1:d}.db", rows, cols)
        Return File.Exists(filename)
    End Function
    Private Sub SaveToFile()
        Dim filename As String = String.Format("puzzle_{0:d}x{1:d}.db", rows, cols)
        Using fs As FileStream = File.OpenWrite(filename)
            Using writer As New BinaryWriter(fs)
                For dbi As Integer = 0 To parts - 1      ' Put count      
                    writer.Write(database(dbi).Count)      ' Write pairs      
                    For Each pair As KeyValuePair(Of UInt32, Integer) In database(dbi)
                        writer.Write(pair.Key)
                        writer.Write(pair.Value)
                    Next
                Next
            End Using
        End Using
    End Sub
    Private Sub LoadFromFile()
        Dim filename As String = String.Format("puzzle_{0:d}x{1:d}.db", rows, cols)
        Using fs As FileStream = File.OpenRead(filename)
            Using reader As New BinaryReader(fs)
                For dbi As Integer = 0 To parts - 1      ' Get count      
                    Dim count As Integer = reader.ReadInt32()      ' Read in all pairs 
                    For i As Integer = 0 To count - 1
                        Dim key As UInt32 = reader.ReadUInt32()
                        Dim value As Integer = reader.ReadInt32()
                        database(dbi).Add(key, value)
                    Next
                Next
            End Using
        End Using
    End Sub

    Public Function GetEstimate(ByVal board As Board) As Integer
        Dim tile As Integer, partition As Integer
        Dim index As UInt32 = 0
        Dim manhattan As Integer = 0
        For i As Integer = 0 To est_hash.Length - 1
            est_hash(i) = 0
        Next
        For row As Integer = 0 To board.Size - 1
            Dim col As Integer = 0
            While col < board.Size
                tile = board.Tiles(row, col) - 1
                partition = partitioning(tile)
                If partition < parts Then   ' if it is not the empty tile (empty tile must be omitted from heuristics)   
                    est_hash(partition) = est_hash(partition) Or index << (partitions(partition)(tile) * BITS_PER_POS)
                    manhattan += Math.Abs(row - Int(tile / board.Size)) + Math.Abs(col - (tile Mod board.Size))
                End If
                col += 1
                index += 1
            End While
        Next
        Dim estimate As Integer = 0
        For i As Integer = 0 To est_hash.Length - 1
            If database(i).ContainsKey(est_hash(i)) Then
                estimate += database(i)(est_hash(i))
            End If
        Next
        ' note: 3x3 puzzles allways use `manhattan`, because `estimate` is always 0, because of the empty database   
        ' pattern db estimate is always at least as high as manhattan distance, but the pattern db estimate mustn't add the empty tile   
        ' -> that's why sometimes the manhattan distance might be slightly higher 
        Return Math.Max(estimate, manhattan)
    End Function
    Private Sub InitDatabase()
        Select Case type
            Case PuzzleType.PUZZLE_3x3
                parts = 8
                rows = 3 : cols = 3
                partitioning = New Integer() {0, 1, 2, 3, 4, 5, 6, 7, 8}
                partitions = New Integer()() {New Integer() {0, -1, -1, -1, -1, -1, -1, -1, 1}, _
                                              New Integer() {-1, 0, -1, -1, -1, -1, -1, -1, 1}, _
                                              New Integer() {-1, -1, 0, -1, -1, -1, -1, -1, 1}, _
                                              New Integer() {-1, -1, -1, 0, -1, -1, -1, -1, 1}, _
                                              New Integer() {-1, -1, -1, -1, 0, -1, -1, -1, 1}, _
                                              New Integer() {-1, -1, -1, -1, -1, 0, -1, -1, 1}, _
                                              New Integer() {-1, -1, -1, -1, -1, -1, 0, -1, 1}, _
                                              New Integer() {-1, -1, -1, -1, -1, -1, -1, 0, 1}}
            Case PuzzleType.PUZZLE_4x4
                parts = 3     ' 5-5-5     
                rows = 4 : cols = 4
                partitioning = New Integer() {0, 0, 0, 0, 1, 1, 0, 2, 1, 1, 2, 2, 1, 2, 2, 3}
                partitions = New Integer()() {New Integer() {0, 1, 2, 3, -1, -1, 4, -1, -1, -1, -1, -1, -1, -1, -1, 5}, _
                                              New Integer() {-1, -1, -1, -1, 0, 1, -1, -1, 2, 3, -1, -1, 4, -1, -1, 5}, _
                                              New Integer() {-1, -1, -1, -1, -1, -1, -1, 0, -1, -1, 1, 2, -1, 3, 4, 5}}
            Case Else
                Throw New Exception("Invalid PuzzleType! Supported types are 3x3 and 4x4.")
        End Select
        database = New Dictionary(Of UInt32, Integer)(parts - 1) {}
        For p As Integer = 0 To parts - 1
            database(p) = New Dictionary(Of UInt32, Integer)()
        Next
        est_hash = New UInt32(parts - 1) {}
    End Sub

    Private Sub BreadthFirstSearchGenerator(ByVal partition As Integer)
        Dim node As New BFSNode(Me), nextNode As BFSNode = Nothing
        Dim counter As Integer
        For p As Integer = 0 To parts - 1
            StoreInDatabase(node, p)
        Next   ' run bfs   
        Dim queue As New Queue(Of BFSNode)()
        queue.Enqueue(node)
        While queue.Count > 0
            node = queue.Dequeue()
            If node.CanGoLeft() Then
                nextNode = node.MoveLeft(partition)
                If StoreInDatabase(nextNode, partition) Then queue.Enqueue(nextNode)
            End If
            If node.CanGoRight() Then
                nextNode = node.MoveRight(partition)
                If StoreInDatabase(nextNode, partition) Then queue.Enqueue(nextNode)
            End If
            If node.CanGoUp() Then
                nextNode = node.MoveUp(partition)
                If StoreInDatabase(nextNode, partition) Then queue.Enqueue(nextNode)
            End If
            If node.CanGoDown() Then
                nextNode = node.MoveDown(partition)
                If StoreInDatabase(nextNode, partition) Then queue.Enqueue(nextNode)
            End If
            If queue.Count > counter Then counter = queue.Count
        End While
        Debug.Print(counter)
    End Sub
    Private Function StoreInDatabase(ByVal node As BFSNode, ByVal partition As Integer) As Boolean
        Dim key As UInt32 = node.GetHashKey(partition)
        If Not database(partition).ContainsKey(key) Then
            database(partition).Add(key, node.distance)
            Return True
        End If
        Return False
    End Function
    Private Sub CleanDatabase(ByVal partition As Integer)
        Dim key As UInt32, mask As UInt32 = Not (BIT_MASK << (partitions(partition)((rows * cols) - 1) * BITS_PER_POS))
        Dim db As New Dictionary(Of UInt32, Integer)()
        For Each pair As KeyValuePair(Of UInt32, Integer) In database(partition)    ' cancel the empty tile    
            key = CType(pair.Key, UInt32) And mask    ' remove the empty tile position from the hash    
            If Not db.ContainsKey(key) Then
                db.Add(key, pair.Value)
            ElseIf db(key) > pair.Value Then
                db(key) = pair.Value
            End If
        Next
        database(partition) = db   ' overwrite db  
    End Sub

    Private Class BFSNode
        Private db As PatternDatabase
        Private board As MiniBoard
        Private empty_row, empty_col, m_distance As Integer
        Private direction As Direction
        Public Property distance() As Integer
            Get
                Return m_distance
            End Get
            Set(ByVal value As Integer)
                m_distance = value
            End Set
        End Property
        Public Sub New(ByVal db As PatternDatabase)
            ' initial state --> the empty tile is in the bottom right corner 
            Me.db = db
            Me.empty_row = db.rows - 1
            Me.empty_col = db.cols - 1
            Me.direction = direction.NONE
            Me.board = New MiniBoard(db.rows, db.cols)
            Me.distance = 0
        End Sub
        Private Sub New(ByVal db As PatternDatabase, ByVal board As MiniBoard, ByVal empty_row As Integer, ByVal empty_col As Integer, ByVal distance As Integer, ByVal direction As Direction)
            Me.db = db
            Me.empty_row = empty_row
            Me.empty_col = empty_col
            Me.board = board
            Me.distance = distance
            Me.direction = direction
        End Sub
        Public Function GetHashKey(ByVal partition As Integer) As UInt32
            Dim hash As UInt32 = 0
            Dim pos As Integer
            Dim index As Integer = 0, im As Integer = db.rows * db.cols
            While index < im
                pos = db.partitions(partition)(board.GetTile(index))
                If pos >= 0 Then      ' correct partition? (including the empty tile)
                    ' then save the current index to the default tile position    
                    hash = hash Or CType(index, UInt32) << (pos * BITS_PER_POS)
                End If
                index += 1
            End While
            Return hash
        End Function
        Public Function CanGoLeft() As Boolean
            Return ((empty_col > 0) AndAlso (direction <> direction.RIGHT))
        End Function
        Public Function CanGoRight() As Boolean
            Return ((empty_col < (db.cols - 1)) AndAlso (direction <> direction.LEFT))
        End Function
        Public Function CanGoUp() As Boolean
            Return ((empty_row > 0) AndAlso (direction <> direction.DOWN))
        End Function
        Public Function CanGoDown() As Boolean
            Return ((empty_row < (db.rows - 1)) AndAlso (direction <> direction.UP))
        End Function
        Public Function MoveLeft(ByVal partition As Integer) As BFSNode
            Dim b As New MiniBoard(board)
            Dim index As Integer = (empty_row * db.cols) + empty_col
            Dim tile As UInt64 = b.GetTile(index - 1)
            b.SetTile(index - 1, b.GetTile(index))
            b.settile(index, tile)
            Dim dist As Integer = distance + (If((db.partitioning(tile) = partition), 1, 0))
            Return New BFSNode(db, b, empty_row, empty_col - 1, dist, direction.LEFT)
        End Function
        Public Function MoveRight(ByVal partition As Integer) As BFSNode
            Dim b As New MiniBoard(board)
            Dim index As Integer = (empty_row * db.cols) + empty_col
            Dim tile As UInt64 = b.GetTile(index + 1)
            b.SetTile(index + 1, b.GetTile(index))
            b.SetTile(index, tile)
            Dim dist As Integer = distance + (If((db.partitioning(tile) = partition), 1, 0))
            Return New BFSNode(db, b, empty_row, empty_col + 1, dist, direction.RIGHT)
        End Function
        Public Function MoveUp(ByVal partition As Integer) As BFSNode
            Dim b As New MiniBoard(board)
            Dim index As Integer = (empty_row * db.cols) + empty_col
            Dim tile As UInt64 = b.GetTile(index - db.cols)
            b.SetTile(index - db.cols, b.GetTile(index))
            b.SetTile(index, tile)
            Dim dist As Integer = distance + (If((db.partitioning(tile) = partition), 1, 0))
            Return New BFSNode(db, b, empty_row - 1, empty_col, dist, direction.UP)
        End Function
        Public Function MoveDown(ByVal partition As Integer) As BFSNode
            Dim b As New MiniBoard(board)
            Dim index As Integer = (empty_row * db.cols) + empty_col
            Dim tile As UInt64 = b.GetTile(index + db.cols)
            b.SetTile(index + db.cols, b.GetTile(index))
            b.SetTile(index, tile)
            Dim dist As Integer = distance + (If((db.partitioning(tile) = partition), 1, 0))
            Return New BFSNode(db, b, empty_row + 1, empty_col, dist, direction.DOWN)
        End Function
    End Class
End Class
