﻿Imports System
Imports System.Collections.Generic
Imports System.Linq
Imports System.Text
Imports System.Diagnostics

Public Class PuzzleSolver
    Private patternDB As PatternDatabase
    Private board_size As Integer

    Public Sub New(ByVal board_size As Integer)
        Me.board_size = board_size
        Dim type As PuzzleType
        Select Case board_size
            Case 3 : type = PuzzleType.PUZZLE_3x3
            Case 4 : type = PuzzleType.PUZZLE_4x4
            Case Else
                Throw New Exception("Unsupported board dimensions! Supported are 3x3 and 4x4 puzzles!")
        End Select
        patternDB = New PatternDatabase(type)
    End Sub

    Public Function Solve(ByVal strConfig As String) As Solution
        Dim arrI As Integer(), arrS As String()
        arrS = strConfig.Split(","c)
        ReDim arrI(arrS.Length - 1)
        For i As Integer = 0 To arrS.Length - 1
            If Not Integer.TryParse(arrS(i), arrI(i)) Then Return Nothing
        Next
        Return Solve(arrI)
    End Function
    Public Function Solve(ByVal board_config As Integer()) As Solution
        Debug.Assert(board_config IsNot Nothing)
        Dim cfg As Integer(,) = New Integer(board_size - 1, board_size - 1) {}
        Dim idx As Integer = board_config.GetLowerBound(0)
        For i As Integer = 0 To board_size - 1
            For j As Integer = 0 To board_size - 1
                cfg(i, j) = board_config(idx)
                idx += 1
            Next
        Next
        Dim board As New Board(cfg)
        Return IDAStar(board)
    End Function
    ' Algorithm IDA* 
    Private Function IDAStar(ByVal board As Board) As Solution
        Dim solution As New Solution(Nothing, Heuristic(board))
        Dim counter As Integer
        While True
            solution = DLS(0, board, solution)
            If (solution.solved = True) OrElse (solution.cost = Integer.MaxValue) Then    ' either solved or cannot be solved? yes->stop   
                Return solution
            End If
            counter += 1
            If counter > 10 Then Exit While
        End While
        Return Nothing
    End Function
    ' Algorithm Depth Limited Search ' # returns (move-sequence or None, new cost limit) 
    Private Function DLS(ByVal start_cost As Integer, ByVal board As Board, ByVal solution As Solution) As Solution
        Dim h As Integer = Heuristic(board)
        Dim minimum_cost As Integer = start_cost + h
        If minimum_cost > solution.cost Then
            Return New Solution(Nothing, minimum_cost)
        End If
        If h = 0 Then
            solution.solved = True
            Return solution
        End If  ' `h` must be 0 in the final state!! (+ it is faster then check each tile using the `IsGoal` method)  
        Dim next_cost_limit As Integer = Integer.MaxValue
        For Each m As Direction In board.Successors(solution.path)
            Dim new_start_cost As Integer = start_cost + board.MoveCost(m)
            solution.path = MoveTowards(board, solution.path, m)
            Dim new_solution As Solution = DLS(new_start_cost, board, solution)
            If new_solution.solved = True Then
                Return new_solution
            End If
            solution.path = MoveBackwards(board, solution.path)
            next_cost_limit = Math.Min(next_cost_limit, new_solution.cost)
        Next
        Return New Solution(Nothing, next_cost_limit)
    End Function
    Private Function Heuristic(ByVal board As Board) As Integer
        'return MisplacedTilesHeuristic(board)
        'return ManhattanDistanceHeuristic(board)
        'return ManhattanDistanceWithLinearConflictHeuristic(board)
        Return PatternDatabaseHeuristic(board)
    End Function
    Private Function PatternDatabaseHeuristic(ByVal board As Board) As Integer
        ' Statically-Partitioned Additive Pattern Database Heuristic  
        Return patternDB.GetEstimate(board)
    End Function
    Private Function MisplacedTilesHeuristic(ByVal board As Board) As Integer
        Dim h As Integer
        For i As Integer = 0 To board.Size - 1
            For j As Integer = 0 To board.Size - 1
                If board.Tiles(i, j) <> ((i * board.Size) + (j + 1)) Then
                    h += 1
                End If
            Next
        Next
        Dim empty_tile As Integer = (board.Size * board.Size) - 1
        If board.Tiles(board.Size - 1, board.Size - 1) <> empty_tile Then   ' empty tile must be omitted from heuristics   
            h -= 1
        End If
        Return h
    End Function
    Private Function ManhattanDistanceHeuristic(ByVal board As Board) As Integer
        Dim h, row, col As Integer
        Dim empty_tile As Integer = (board.Size * board.Size) - 1
        For i As Integer = 0 To board.Size - 1
            For j As Integer = 0 To board.Size - 1
                If board.Tiles(i, j) = empty_tile Then
                    Continue For
                End If    ' empty tile must be omitted from heuristics    
                row = Int((board.Tiles(i, j) - 1) / board.Size)
                col = (board.Tiles(i, j) - 1) Mod board.Size
                h += Math.Abs(row - i) + Math.Abs(col - j)
            Next
        Next
        Return h
    End Function
    Private Function ManhattanDistanceWithLinearConflictHeuristic(ByVal board As Board) As Integer
        Dim h As Integer = ManhattanDistanceHeuristic(board)
        Dim row As Integer, col As Integer
        Dim empty_tile As Integer = (board.Size * board.Size) - 1  '   --> in rows  
        For i As Integer = 0 To board.Size - 1
            For j As Integer = 0 To board.Size - 1
                If board.Tiles(i, j) = empty_tile Then
                    Continue For
                End If    ' empty tile must be omitted from heuristics    '    
                row = Int((board.Tiles(i, j) - 1) / board.Size)
                If row = i Then     ' is the tile in it's correct row?     
                    ' is there any [i,k] tile from the same row that needs to go before the [i,j] tile?    
                    col = (board.Tiles(i, j) - 1) Mod board.Size
                    For k As Integer = j + 1 To board.Size - 1
                        If Int((board.Tiles(i, k) - 1) / board.Size) = i Then         ' the same row?     
                            If ((board.Tiles(i, k) - 1) Mod board.Size) < col Then ' before the [i,j] tile?   
                                h += 2
                            End If
                        End If
                    Next
                End If
            Next
        Next  '   --> in columns
        For j As Integer = 0 To board.Size - 1
            For i As Integer = 0 To board.Size - 1
                If board.Tiles(i, j) = empty_tile Then
                    Continue For
                End If    ' empty tile must be omitted from heuristics    '  
                col = (board.Tiles(i, j) - 1) Mod board.Size
                If col = j Then     ' is the tile in it's correct column?     
                    ' is there any [k,j] tile from the same column that needs to go before the [i,j] tile?  
                    row = Int((board.Tiles(i, j) - 1) / board.Size)
                    For k As Integer = i + 1 To board.Size - 1
                        If ((board.Tiles(k, j) - 1) Mod board.Size) = j Then       ' the same column?      
                            If Int((board.Tiles(k, j) - 1) / board.Size) < row Then   ' before the [k,j] tile?        
                                h += 2
                            End If
                        End If
                    Next
                End If
            Next
        Next  '  
        Return h
    End Function
    Private Function MoveTowards(ByVal board As Board, ByVal path As List(Of Direction), ByVal move As Direction) As List(Of Direction) ' Expand  
        If path Is Nothing Then path = New List(Of Direction)()
        board.Move(move)
        path.Add(move)
        Return path
    End Function
    Private Function MoveBackwards(ByVal board As Board, ByVal path As List(Of Direction)) As List(Of Direction) ' Rollback  
        Dim reverse_direction As Direction() = {Direction.NONE, Direction.RIGHT, Direction.LEFT, Direction.DOWN, Direction.UP}
        If path Is Nothing Then Return Nothing
        board.Move(reverse_direction(CInt(path(path.Count - 1))))
        path.RemoveAt(path.Count - 1)
        Return path
    End Function
End Class

Public Class Solution
    Private m_cost As Integer
    Private m_path As List(Of Direction)
    Private m_solved As Boolean
    Public Property cost() As Integer
        Get
            Return m_cost
        End Get
        Set(ByVal value As Integer)
            m_cost = value
        End Set
    End Property
    Public Property path() As List(Of Direction)
        Get
            Return m_path
        End Get
        Set(ByVal value As List(Of Direction))
            m_path = value
        End Set
    End Property
    Public Property solved() As Boolean
        Get
            Return m_solved
        End Get
        Set(ByVal value As Boolean)
            m_solved = value
        End Set
    End Property
    Public Sub New(ByVal path As List(Of Direction), ByVal cost As Integer)
        Me.cost = cost
        Me.path = path
        Me.solved = False
    End Sub
    Public Function PrintPath() As String
        Dim sb As New StringBuilder
        If path Is Nothing Then
            sb.Append("Solution path is empty - no moves!")
        Else
            sb.AppendFormat("Solution path ({0:d}): ", path.Count)
            sb.AppendLine()
            For Each d As Direction In path
                sb.AppendLine(d.ToString() & ",")
            Next
        End If
        Return sb.ToString
    End Function
End Class

Public Enum Direction
    NONE
    LEFT
    RIGHT
    UP
    DOWN
End Enum
Public Enum PuzzleType
    PUZZLE_3x3
    PUZZLE_4x4
End Enum